import os
import sys
import json

import click

from pypubmed.core.eutils import Eutils
from pypubmed.core.citations import ncbi_citations, manual_citations


__epilog__ = '''
examples:\n
    citations 1 2 3\n
    citations 1 2 3 -f nlm\n
    citations 1 2 3 -m -f apa\n
'''

@click.command(name='citations', epilog=__epilog__, help=click.style('generate citations for given pmids', bold=True, fg='magenta'), no_args_is_help=True)
@click.option('-m', '--manual', help='cite with manual citations, default with ncbi citations', default=False, is_flag=True)
@click.option('-f', '--fmt', help='the format of citation', type=click.Choice('ama mla apa nlm'.split()), default='ama')
@click.option('-o', '--outfile', help='the output filename [stdout]')
@click.argument('pmids', nargs=-1)
@click.pass_obj
def citations_cli(obj, **kwargs):
    pmids = kwargs['pmids']
    if not pmids:
        click.secho('please supply pmids or a file', fg='green')
        exit(1)

    if os.path.isfile(pmids[0]):
        pmid_list = open(pmids[0]).read().strip().split()
    else:
        pmid_list = pmids
    
    out = open(kwargs['outfile'], 'w') if kwargs['outfile'] else sys.stdout
    with out:
        if kwargs['manual']:
            e = Eutils(api_key=obj['api_key'])
            articles = e.efetch(pmid_list)

            click.echo('\n', err=True)

            for article in articles:
                citation = manual_citations(article, fmt=kwargs['fmt'])
                out.write('{}\t{}\n'.format(article.pmid, citation))
        else:
            for pmid in pmid_list:
                citation = ncbi_citations(pmid, fmt=kwargs['fmt'])['orig']
                out.write('{}\t{}\n'.format(pmid, citation))



import os
import re
import click
import datetime
import pickle

from dateutil.parser import parse as date_parse

from pypubmed.util import safe_open
from pypubmed.core.export import Export

search_examples = click.style('''
examples:

    pypubmed search ngs -l 5 -o ngs.xlsx

    pypubmed search 'NGS[Title] AND Disease[Title/Abstract]' -o ngs_disease.xlsx

    pypubmed search 1,2,3,4

    pypubmed search pmid_list.txt
''', fg='yellow')
@click.command(help=click.style('search with pmid or a term', bold=True, fg='green'), epilog=search_examples, no_args_is_help=True)
@click.option('-cit', '--cited', help='get cited information', default=False, is_flag=True)
@click.option('-n', '--no-translate', help='do not translate the abstract', default=False, is_flag=True)
@click.option('-b', '--batch-size', help='the batch size for efetch', default=10, type=int, show_default=True)
@click.option('-min', '--min-factor', help='filter with IF', type=float)
@click.option('-l', '--limit', help='limit the count of output', type=int)
@click.option('-f', '--fields', help='the fields to export')
@click.option('-o', '--outfile', help='the output filename', default='pubmed.xlsx', show_default=True)
@click.option('-a', '--author', help='export information of authors', is_flag=True, hidden=True)

@click.option('-c', '--cache', help='store translated result to a cache file', is_flag=True)
@click.option('-s', '--retstart', help='the number of start', type=int, default=0, show_default=True)

@click.argument('term', nargs=1)
@click.pass_obj
def search(obj, **kwargs):
    
    data = []
    translate_cache = {}
    cache_file = '.translate.cache.pkl'
    if os.path.isfile(cache_file):
        with safe_open(cache_file, 'rb') as f:
            translate_cache = pickle.load(f)

    articles = obj['eutils'].search(translate=not kwargs['no_translate'], translate_cache=translate_cache, **kwargs)

    try:
        for n, article in enumerate(articles, 1):
            obj['eutils'].logger.debug(f'{n}. {article}')

            # store translated result to cache file
            if kwargs['cache']:
                translate_cache[article.pmid] = article.abstract_cn

            # filter impact factor
            if kwargs['min_factor'] and (article.impact_factor != '.' and article.impact_factor < kwargs['min_factor']):
                continue

            # export author information or not
            if not kwargs['author']:
                del article.author_mail
                del article.author_first
                del article.author_last

            data.append(article.to_dict())

            if kwargs['limit'] and n >= kwargs['limit']:
                break
    except KeyboardInterrupt:
        pass

    if kwargs['min_factor']:
        obj['eutils'].logger.info('A total of {} articles were found, {} remaining after filtering IF>={}'.format(n, len(data), kwargs['min_factor']))

    if kwargs['cache']:
        with safe_open(cache_file, 'wb') as out:
            pickle.dump(translate_cache, out)
        obj['eutils'].logger.debug(f'save cache file: {cache_file}')

    Export(data, **kwargs).export()


@click.command(help=click.style('generate advance search string', bold=True, fg='cyan'))
@click.pass_obj
def advance_search(obj, **kwargs):
    fields = obj['eutils'].validate_fields()

    query_box = ''
    while True:
        number = click.prompt('>>> please choose a number of field', type=int, default=48)
        field = fields[number][1]
        click.secho(f'your choice is: {number} - {field}', fg='cyan')

        if field.startswith('Date - '):
            start = click.prompt('>>> please enter the start date(YYYY-MM-DD)', value_proc=date_parse).strftime('%Y/%m/%d')
            end = click.prompt('>>> please enter the end date', default='3000', value_proc=date_parse)
            if end != '3000':
                end = end.strftime('%Y/%m/%d')
            term = f'("{start}"[{field}] : "{end}"[{field}])'
        else:
            term = click.prompt('>>> please enter a search term')
            if field == 'All Fields':
                term = f'{term}'
            else:
                term = f'"{term}"[{field}]'

        if not query_box:
            query_box = term
        else:
            logic = click.prompt('>>> please input the logic', type=click.Choice(['and', 'or', 'not']), default='and').upper()
            query_box = f'({query_box}) {logic} ({term})'

        click.secho('query box now: ' + query_box, fg='bright_green')

        if click.confirm('input finish?'):
            break

    click.secho(f'final query box: {query_box}', fg='bright_cyan')
    res = obj['eutils'].esearch(query_box, retmax=1, head=True)

    if res['count']:
        details = []
        for each in res['translationstack']:
            if isinstance(each, dict):
                details += ['{term}:{count}'.format(**each)]
        click.secho('count:\t{count}\nquery:\t{querytranslation}\ndetail:\t{detail}'.format(detail=', '.join(details), **res), fg='yellow')

        if click.confirm('search with this query box?'):
            outfile = re.sub(r'[\'"\(\)\[\]/ ]+', '_', query_box).strip('_') + '.xlsx'
            outfile = click.prompt('output filename', default=outfile, show_default=True)
            os.system(f'pypubmed search "{query_box}" -o {outfile}')
    else:
        click.echo(res)



"""
    https://www.ncbi.nlm.nih.gov/books/NBK25497/
"""
import os
import sys
import json
import time
import textwrap
import datetime

import click
import prettytable

from dateutil.parser import parse as date_parse

from impact_factor.core import Factor as ImpactFactor
from googletranslatepy import Translator as GoogleTrans
from simple_loggers import SimpleLogger
from webrequests import WebRequest


from pypubmed.util import xml_parser
from pypubmed.core.article import Article


class Eutils(object):
    """
        params:
            db          database name
            api_key     api_key or NCBI_API_KEY in environment

        optional params:
            term        term for esearch
            id          id(s) for efetch
            field       field for esearch
    """
    base_url = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/'
    logger = SimpleLogger('Eutils')
    IF = ImpactFactor()

    def __init__(self, db='pubmed', proxies=None, api_key=None, **kwargs):
        self.db = db
        self.api_key = api_key
        self.validate_api_key()
        self.TR = GoogleTrans(proxies=proxies)

        self.TR_OK = self.TR.check_proxies()


    def parse_params(self, **kwargs):
        """
            - add default db
            - add api_key if available
        """
        params = {'db': self.db}
        if self.api_key:
            params['api_key'] = self.api_key
        params.update(kwargs)
        
        if 'api_key' in params and params['api_key'] is None:
            del params['api_key']

        return params

    def esearch(self, term, retstart=0, retmax=250, head=False, limit=None, **kwargs):
        """
            https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.ESearch

            > -search from a database from given term
            >> - esearch.cgi?db=pubmed&term=ngs
            >> - esearch.cgi?db=pubmed&term=ngs&retmode=xml&field=TIAB
            >> - esearch.cgi?db=pubmed&term=ngs[Title/Abstract]&retmode=xml
        """
        url = self.base_url + 'esearch.fcgi'
        params = self.parse_params(term=term, retmode='json', retstart=retstart, retmax=retmax, **kwargs)

        # print(params)

        result = WebRequest.get_response(url, params=params).json()['esearchresult']

        if head:
            return result

        self.logger.info('{count} articles found with term: {querytranslation}'.format(**result))

        if limit is None and int(result['count']) > 250:
            self.logger.warning('too many results, you can limit output with option "-l/--limit N", '
                                'or simplify your input with sub-command "advance-search" ')
            exit(1)

        idlist = result['idlist']

        while int(result['retstart']) + int(result['retmax']) < int(result['count']):
            if limit and len(idlist) >= limit:
                break
            retstart = int(result['retstart']) + int(result['retmax'])
            params = self.parse_params(term=term, retmode='json', retstart=retstart, retmax=retmax, **kwargs)
            result = WebRequest.get_response(url, params=params).json()['esearchresult']
            idlist += result['idlist']

        if limit:
            self.logger.info('limit {} from {}'.format(limit, result['count']))
            idlist = idlist[:limit]

        if idlist:
            self.logger.debug('idlist: {} ...'.format(', '.join(idlist[:10])))
        else:
            self.logger.warning('no result for term: {}'.format(term))

        return idlist

    def efetch(self, ids, batch_size=5, **kwargs):
        """
            https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EFetch

            > - fetch from a database for given ids
            >> efetch.cgi?db=pubmed&id=1,2,3
        """
        url = self.base_url + 'efetch.fcgi'

        self.logger.info('fetching start: total {}, batch_size: {}'.format(len(ids), batch_size))

        for n in range(0, len(ids), batch_size):
            _id = ','.join(ids[n:n+batch_size])

            self.logger.debug(f'fetching xml: {n+1} - {n+batch_size}')
            params = self.parse_params(id=_id, retmode='xml')
            xml = WebRequest.get_response(url, params=params).text
            
            self.logger.debug(f'parsing xml: {n+1} - {n+batch_size}')
            for context in xml_parser.parse(xml):
                article = Article(**context)
                yield article

    def einfo(self, **kwargs):
        """
            https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EInfo

            > - show all database list
            >> einfo.fcgi?db=
            > - show dbinfo for given database
            >> einfo.fcgi?db=pubmed
        """
        url = self.base_url + 'einfo.fcgi'
        params = self.parse_params(retmode='json', **kwargs)
        info = WebRequest.get_response(url, params=params, allowed_codes=[200, 400]).json()
        return info

    def elink(self, ids, dbfrom='pubmed', cmd='neighbor', **kwargs):
        """
            https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.ELink

            > - get cited (`{"linkname": "pubmed_pubmed_citedin"}`)
            >> elink.fcgi?dbfrom=pubmed&db=pubmed&id=20210808&cmd=neighbor&retmode=json
            > - get pdf url
            >> elink.fcgi?dbfrom=pubmed&db=pubmed&cmd=prlinks&id=10210801

            cmds:
                - neighbor (default)
                - neighbor_score
                - neighbor_history
                - acheck
                - ncheck
                - lcheck
                - llinks
                - llinkslib
                - prlinks
        """
        url = self.base_url + 'elink.fcgi'
        params = self.parse_params(retmode='json', id=ids, dbfrom=dbfrom, cmd=cmd,  **kwargs)
        result = WebRequest.get_response(url, params=params).json()
        return result

    def get_cited(self, _id, dbfrom='pubmed', cmd='neighbor'):
        """
            get the cited pmids for given pmid
        """
        linksetdbs = self.elink(_id, dbfrom=dbfrom, cmd=cmd)['linksets'][0]['linksetdbs']
        links = [
            linkset['links'] for linkset in linksetdbs
            if linkset['linkname'] == 'pubmed_pubmed_citedin'
        ]
        
        citedin = links[0] if links else []
        return {'count': len(citedin), 'links': citedin}

    def get_pdf_url(self, _id, dbfrom='pubmed', cmd='prlinks'):
        """
            get the pdf url for given pmid
        """
        idurllist = self.elink(_id, dbfrom=dbfrom, cmd=cmd)['linksets'][0]['idurllist']
 
        result = {each['id']: each['objurls'][0]['url']['value'] for each in idurllist}

        return result

    def validate_fields(self, **kwargs):
        """
            Show all fields for given database
        """
        fieldlist = self.einfo(**kwargs)['einforesult']['dbinfo'][0]['fieldlist']

        fieldlist = sorted(fieldlist, key=lambda x: x.get('fullname'))

        fields = {}
        table = prettytable.PrettyTable(field_names=['Number', 'Name', 'FullName', 'Description'])
        for n, each in enumerate(fieldlist, 1):
            # print('{n}\t{name}\t{fullname}\t{description}'.format(n=n, **each))
            fields[n] = [each['name'], each['fullname'], each['description']]
            table.add_row([n, each['name'], each['fullname'], each['description']])

        table.align['FullName'] = 'l'
        table.align['Description'] = 'l'
        click.secho(str(table), fg='bright_blue')

        return fields

    def validate_api_key(self):
        """
            Description: https://www.ncbi.nlm.nih.gov/account/settings/#accountSettingsApiKeyManagement
                - E-utils users are allowed 3 requests/second without an API key.
                - Create an API key to increase your e-utils limit to 10 requests/second.
        """
        configfile = os.path.join(os.path.expanduser('~'), '.pypubmed.cfg')

        if not self.api_key:
            if os.path.isfile(configfile):
                self.api_key = open(configfile).read().strip()
            if not self.api_key:
                msg = textwrap.dedent('''
                    API_KEY not found! Use eutils with a api_key is a good idea!
                    You can create one from: https://www.ncbi.nlm.nih.gov/account/settings/#accountSettingsApiKeyManagement
                    If you already have one, you can add it to your environment: export NCBI_API_KEY=xxxx''')
                self.logger.warning(msg)
                return

        res = self.einfo()
        if 'error' in res:
            self.logger.warning('invalid api_key, please check: {}'.format(self.api_key))
            self.api_key = None
        else:
            self.logger.info('Valid api_key: {}'.format(self.api_key))
            with open(configfile, 'w') as out:
                out.write(self.api_key)

    def search(self, term, cited=True, translate=True, impact_factor=True, translate_cache=None, **kwargs):
        """
            term:
                - string, eg. 'ngs AND disease'
                - pmid, eg. 1,2,3
                - file with pmid
        """
        if os.path.isfile(term):
            idlist = open(term).read().strip().split()
        elif all(each.isdigit() for each in term.split(',')):
            idlist = term.split(',')
        else:
            idlist = self.esearch(term, **kwargs)

        articles = self.efetch(idlist, **kwargs)
        for article in articles:
            if impact_factor:
                res = self.IF.search(article.issn) or self.IF.search(article.e_issn)
                article.impact_factor = res[0]['factor'] if res else '.'

            if cited:
                article.cited = self.get_cited(article.pmid)

            if translate and self.TR_OK:
                if translate_cache and translate_cache.get(article.pmid):
                    article.abstract_cn = translate_cache.get(article.pmid)
                else:
                    try:
                        article.abstract_cn = self.TR.translate(article.abstract)
                    except Exception as e:
                        article.abstract_cn = 'translate failed'
            yield article



import os
import sys
import logging

import click

from ._search import search, advance_search
from ._citations import citations_cli
from pypubmed import version_info
from pypubmed.core.eutils import Eutils


log_level_maps = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
}

__epilog__ = click.style('contact: {author} <{author_email}>'.format(**version_info), fg='bright_black')

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(context_settings=CONTEXT_SETTINGS, epilog=__epilog__, help=click.style(version_info['desc'], fg='bright_blue', bold=True))
@click.option('-l', '--log-level', help='the mode of loggging',
              show_default=True, default='debug',
              type=click.Choice(log_level_maps.keys()))
@click.option('-k', '--api-key', help='the api_key of NCBI Pubmed, NCBI_API_KEY environment is available',
              envvar='NCBI_API_KEY', show_envvar=True)
@click.option('-p', '--proxies', help='the proxies for google translate, eg. http://127.0.0.1:1080',)
@click.version_option(version=version_info['version'], prog_name='pypubmed')
@click.pass_context
def cli(ctx, **kwargs):

    kwargs['log_level'] = log_level_maps[kwargs['log_level']]
    e = Eutils(api_key=kwargs['api_key'], proxies=kwargs['proxies'])
    e.logger.level = int(kwargs['log_level'])
    ctx.obj = {
        'eutils': e
    }


def main():
    cli.add_command(search)
    cli.add_command(advance_search)
    cli.add_command(citations_cli)
    cli()


if __name__ == '__main__':
    main()
